"""
"""


#[
from __future__ import annotations

from typing import (Any, Iterable, Sequence, Literal, Self, Protocol, NoReturn, )
from types import (EllipsisType, )
from numbers import (Real, )
from collections import (namedtuple, )
import functools as _ft
import dataclasses as _dc
import numpy as _np
import scipy as _sp
import wlogging as _wl

from .. import has_variants as _has_variants
from .. import quantities as _quantities
from .. import pages as _pages
from ..frames import (Frame, )
from ..dataslates.main import (Dataslate, )
from ..databoxes.main import (Databox, )
from .solutions import (Solution, right_div, left_div, )
from ..dates import (Dater, )

from . import initializers as _initializers
from . import shock_simulators as _shock_simulators
from .descriptors import (Squid, )
#]


_simulate_anticipated_shocks \
    = _shock_simulators.simulate_triangular_anticipated_shocks


_DEFAULT_DELTA_TOLERANCE = 1e-12
_LOGGER = _wl.get_colored_two_liner(__name__, level=_wl.INFO, )


class SingularMatrixError(ValueError, ):
    pass


class KalmanFilterableProtocol(Protocol, ):
    """
    """
    #[

    is_singleton: bool

    def get_variant(self, vid: int, /) -> Self: ...

    def get_singleton_solution(self, deviation: bool = False, ) -> Solution: ...

    def get_cov_unanticipated_shocks(self, ) -> _np.ndarray: ...

    #]


class KalmanOutputData:
    """
......................................................................

==Output data from Kalman filter==


......................................................................
    """
    #[

    __slots__ = (
        "predict_mse_measurement",
        "predict_med",
        "predict_std",
        "predict_err",
        "update_med",
        "update_std",
        "smooth_med",
        "smooth_std",
    )

    def __init__(
        self: Self,
        output_store: _OutputStore,
    ) -> None:
        """
        """
        for attr_name in self.__slots__:
            attr = getattr(output_store, attr_name, )
            if hasattr(attr, "to_output_arg", ):
                self_attr = attr.to_output_arg()
            else:
                self_attr = attr
            setattr(self, attr_name, self_attr, )

    def __repr__(self, /, ) -> str:
        """
        """
        slots_to_repr = (s for s in self.__slots__ if getattr(self, s) is not None)
        return f"KalmanOutputData({', '.join(slots_to_repr)})"

    #]


class Needs:
    """
    """
    #[

    __slots__ = (
        "return_predict",
        "return_update",
        "return_smooth",
        "return_stds",
        "return_predict_err",
        "return_predict_mse_measurement",
        "rescale_variance",
    )

    def __init__(
        self: Self,
        return_: Iterable[str, ...],
        return_predict: bool = True,
        return_update: bool = True,
        return_smooth: bool = True,
        return_stds: bool = True,
        return_predict_err: bool = True,
        return_predict_mse_measurement: bool = True,
        rescale_variance: bool = False,
    ) -> None:
        """
        """
        return_ = tuple(return_) if not isinstance(return_, str) else (return_, )
        self.return_predict = return_predict and "predict" in return_
        self.return_update = return_update and "update" in return_
        self.return_smooth = return_smooth and "smooth" in return_
        self.return_stds = return_stds and "stds" in return_
        self.return_predict_err = return_predict_err and "predict_err" in return_
        self.return_predict_mse_measurement = return_predict_mse_measurement and "predict_mse_measurement" in return_
        self.rescale_variance = rescale_variance

    #]


class _LogDataslate:
    """
    """
    #[

    @classmethod
    def from_dataslate(
        klass,
        input_ds: Dataslate,
        num_variants: int,
        name_to_log_name: dict[str, str],
    ) -> Self | None:
        """
        """
        self = klass()
        self._dataslate = Dataslate.nan_from_template(input_ds, num_variants=num_variants, )
        self._prepare_log_names(name_to_log_name, )
        return self

    @classmethod
    def from_names_dates(
        klass,
        measurement_names: Iterable[str],
        dates: Iterable[Dater],
        num_variants: int,
        name_to_log_name: dict[str, str],
    ) -> Self:
        """
        """
        self = klass()
        self._dataslate = Dataslate.nan_from_names_periods(
            measurement_names, dates,
            num_variants=num_variants,
        )
        self._prepare_log_names(name_to_log_name, )
        return self

    def store(
        self: Self,
        array: _np.ndarray,
        lhs_indexes: tuple,
        rhs_indexes=None,
        transform=None,
    ) -> None:
        """
        """
        if array is None:
            return
        rhs_indexes = rhs_indexes if rhs_indexes is not None else slice(None, )
        if transform is not None:
            array = transform[rhs_indexes, :] @ array
        else:
            array = array[rhs_indexes]
        self._dataslate._variants[0].data[lhs_indexes] = array

    def _prepare_log_names(
        self,
        name_to_log_name: dict[str, str],
        /,
    ) -> None:
        """
        """
        self._dataslate.rename(name_to_log_name, )
        output_names = self._dataslate.output_names
        self._name_to_log_name = {
            name: log_name
            for name, log_name in name_to_log_name.items()
            if log_name in output_names
        }

    def extend(
        self,
        other: Self,
        /,
    ) -> None:
        """
        """
        self._dataslate.extend(other._dataslate, )

    #]


class _MedLogDataslate(_LogDataslate, ):
    """
    """
    #[

    def to_output_arg(
        self: Self,
        /,
    ) -> Databox:
        """
        """
        db = self._dataslate.to_databox()
        for name, log_name in self._name_to_log_name.items():
            db[name] = db[log_name].copy()
            db[name].exp()
        return db

    #]


class _StdLogDataslate(_LogDataslate, ):
    """
    """
    #[

    def store_from_mse(
        self: Self,
        mse_array: _np.ndarray,
        lhs_indexes: tuple,
        rhs_indexes=None,
        transform=None,
    ) -> None:
        if mse_array is None:
            return
        rhs_indexes = rhs_indexes if rhs_indexes is not None else slice(None, )
        if transform is not None:
            transform_rhs_indexes = transform[rhs_indexes, :]
            mse_array = transform_rhs_indexes @ mse_array @ transform_rhs_indexes.T
        else:
            mse_array = mse_array[rhs_indexes, ...][..., rhs_indexes]
        self._dataslate._variants[0].data[lhs_indexes] = _std_from_mse(mse_array, )

    def to_output_arg(
        self: Self,
        /,
    ) -> Databox:
        """
        """
        return self._dataslate.to_databox()

    #]


class _OutputStore:
    """
    """
    #[

    __slots_to_input__ = (
        "transform",
        "squid",
        "measurement_incidence",
    )

    __slots_to_create__ = (
        "predict_med",
        "predict_std",
        "predict_mse_measurement",

        "update_med",
        "update_std",
        "predict_err",

        "smooth_med",
        "smooth_std",
    )

    __slots__ = __slots_to_input__ + __slots_to_create__

    def __init__(
        self,
        input_ds: Dataslate,
        num_variants: int,
        measurement_names: Iterable[str],
        needs: Needs,
        name_to_log_name: dict[str, str] | None = None,
        **kwargs,
    ) -> Self:
        """
        """
        #
        for n in self.__slots_to_input__:
            setattr(self, n, kwargs.get(n, ), )
        #
        _med_constructor = _ft.partial(
            _MedLogDataslate.from_dataslate,
            input_ds=input_ds,
            num_variants=num_variants,
            name_to_log_name=name_to_log_name,
        )
        #
        _std_constructor = _ft.partial(
            _StdLogDataslate.from_dataslate,
            input_ds=input_ds,
            num_variants=num_variants,
            name_to_log_name=name_to_log_name,
        )
        #
        # Prediction step
        #
        self.predict_med = _med_constructor() if needs.return_predict else None
        self.predict_std = _std_constructor() if needs.return_predict and needs.return_stds else None
        self.predict_mse_measurement = [
            [None]*input_ds.num_periods for _ in range(num_variants, )
        ] if needs.return_predict else None
        #
        # Updating step
        #
        self.update_med = _med_constructor() if needs.return_update else None
        self.update_std = _std_constructor() if needs.return_update and needs.return_stds else None
        self.predict_err = _MedLogDataslate.from_names_dates(
            measurement_names,
            input_ds.periods,
            num_variants,
            name_to_log_name,
        ) if needs.return_predict_err else None
        #
        # Smoothing step
        #
        self.smooth_med = _med_constructor() if needs.return_smooth else None
        self.smooth_std = _std_constructor() if needs.return_smooth and needs.return_stds else None

    def extend(self, other: Self, /, ) -> None:
        """
        """
        for s in self.__slots__:
            if getattr(self, s) is not None:
                getattr(self, s).extend(getattr(other, s), )

    def rescale_stds(self, var_scale, /, ) -> None:
        """
        """
        if var_scale == 1 or var_scale is None:
            return
        std_scale = _sqrt_positive(self.var_scale)
        for attr_name in ("predict_std", "update_std", "smooth_std", ):
            attr = getattr(self, attr_name, )
            if attr is None:
                continue
            attr._variants = [ v.data * std_scale for v in attr._variants ]

    def create_output_data(self: Self, /, ) -> KalmanOutputData:
        """
        """
        return KalmanOutputData(self, )

    def store_predict(
        self: Self,
        t: int | EllipsisType,
        inx_y: _np.ndarray | None = None,
        a0: _np.ndarray | None = None,
        u0: _np.ndarray | None = None,
        v0: _np.ndarray | None = None,
        y0: _np.ndarray | None = None,
        w0: _np.ndarray | None = None,
        Q0: _np.ndarray | None = None,
        F: _np.ndarray | None = None,
        cov_u0: _np.ndarray | None = None,
        cov_w0: _np.ndarray | None = None,
    ) -> None:
        """
        """
        if self.predict_med is not None:
            curr_xi_qids = self.squid.curr_xi_qids
            curr_xi_indexes = self.squid.curr_xi_indexes
            u_qids = self.squid.u_qids
            v_qids = self.squid.v_qids
            w_qids = self.squid.w_qids
            y_qids = self.squid.y_qids
            #
            self.predict_med.store(a0, (curr_xi_qids, t), rhs_indexes=curr_xi_indexes, transform=self.transform, )
            self.predict_med.store(u0, (u_qids, t), )
            self.predict_med.store(v0, (v_qids, t), )
            self.predict_med.store(w0, (w_qids, t), )
            full_y0 = self._expand_y_to_full(y0, t, )
            self.predict_med.store(full_y0, (y_qids, t), )
            #
        if self.predict_std is not None:
            curr_xi_qids = self.squid.curr_xi_qids
            curr_xi_indexes = self.squid.curr_xi_indexes
            std_u_qids = self.squid.std_u_qids
            std_w_qids = self.squid.std_w_qids
            self.predict_std.store_from_mse(Q0, (curr_xi_qids, t), rhs_indexes=curr_xi_indexes, transform=self.transform, )
            self.predict_std.store_from_mse(cov_u0, (std_u_qids, t), )
            self.predict_std.store_from_mse(cov_w0, (std_w_qids, t), )
            #
        if self.predict_mse_measurement is not None and F is not None:
            self.predict_mse_measurement[0][t] = F

    def store_update(
        self: Self,
        t: int | EllipsisType,
        inx_y: _np.ndarray | None = None,
        xi: _np.ndarray | None = None,
        u: _np.ndarray | None = None,
        v: _np.ndarray | None = None,
        w: _np.ndarray | None = None,
        y: _np.ndarray | None = None,
        pe: _np.ndarray | None = None,
        Q: _np.ndarray | None = None,
    ) -> None:
        """
        """
        if self.update_med is not None:
            curr_xi_qids = self.squid.curr_xi_qids
            curr_xi_indexes = self.squid.curr_xi_indexes
            u_qids = self.squid.u_qids
            v_qids = self.squid.v_qids
            w_qids = self.squid.w_qids
            y_qids = self.squid.y_qids
            self.update_med.store(xi, (curr_xi_qids, t), rhs_indexes=curr_xi_indexes, transform=self.transform, )
            self.update_med.store(u, (u_qids, t), )
            self.update_med.store(v, (v_qids, t), )
            self.update_med.store(w, (w_qids, t), )
            full_y = self._expand_y_to_full(y, t, )
            self.update_med.store(full_y, (y_qids, t), )
        if self.update_std is not None:
            curr_xi_qids = self.squid.curr_xi_qids
            curr_xi_indexes = self.squid.curr_xi_indexes
            self.update_std.store_from_mse(Q, (curr_xi_qids, t), rhs_indexes=curr_xi_indexes, transform=self.transform, )
        if self.predict_err is not None:
            full_pe = self._expand_y_to_full(pe, t, )
            self.predict_err.store(full_pe, (..., t), )

    def store_smooth(
        self: Self,
        t: int | EllipsisType,
        inx_y: _np.ndarray | None = None,
        xi: _np.ndarray | None = None,
        u: _np.ndarray | None = None,
        v: _np.ndarray | None = None,
        w: _np.ndarray | None = None,
        y: _np.ndarray | None = None,
        Q: _np.ndarray | None = None,
    ) -> None:
        """
        """
        if self.smooth_med is not None:
            curr_xi_qids = self.squid.curr_xi_qids
            curr_xi_indexes = self.squid.curr_xi_indexes
            u_qids = self.squid.u_qids
            v_qids = self.squid.v_qids
            w_qids = self.squid.w_qids
            y_qids = self.squid.y_qids
            self.smooth_med.store(xi, (curr_xi_qids, t), rhs_indexes=curr_xi_indexes, transform=self.transform, )
            self.smooth_med.store(u, (u_qids, t), )
            self.smooth_med.store(v, (v_qids, t), )
            self.smooth_med.store(w, (w_qids, t), )
            full_y = self._expand_y_to_full(y, t, )
            self.smooth_med.store(full_y, (y_qids, t), )
        if self.smooth_std is not None:
            curr_xi_qids = self.squid.curr_xi_qids
            curr_xi_indexes = self.squid.curr_xi_indexes
            self.smooth_std.store_from_mse(Q, (curr_xi_qids, t), rhs_indexes=curr_xi_indexes, transform=self.transform, )

    def _expand_y_to_full(
        self,
        observed: _np.ndarray | None,
        t: int | EllipsisType,
    ) -> _np.ndarray:
        """
        """
        if observed is None or t is Ellipsis:
            return observed
        inx_y = self.measurement_incidence[:, t]
        full = _np.full(inx_y.shape, _np.nan, )
        full[inx_y] = observed
        return full

    #]


def _sqrt_positive(x):
    return _np.sqrt(_np.maximum(x, 0.0, ))


def _std_from_mse(mse, ):
    return _sqrt_positive(_np.diag(mse).copy())


class Mixin:
    """
    """
    #[

    @_pages.reference(category="filtering", )
    def kalman_filter(
        self,
        #
        input_db: Databox,
        span: Iterable[Dater],
        #
        diffuse_scale: Real | None = None,
        diffuse_method: Literal["approx_diffuse", "fixed_unknown", "fixed_zero", ] = "fixed_unknown",
        #
        return_: Iterable[str, ...] = ("predict", "update", "smooth", "stds", "predict_err", "predict_mse_measurement", ),
        return_predict: bool = True,
        return_update: bool = True,
        return_smooth: bool = True,
        return_stds: bool = True,
        return_predict_err: bool = True,
        return_predict_mse_measurement: bool = True,
        rescale_variance: bool = False,
        #
        shocks_from_data: bool = False,
        stds_from_data: bool = False,
        #
        prepend_initial: bool = False,
        append_terminal: bool = False,
        deviation: bool = False,
        check_singularity: bool = False,
        num_variants: int | None = None,
        #
        unpack_singleton: bool = True,
        logging_level: int = _wl.INFO,
    ) -> Databox:
        r"""
················································································

==Run Kalman filter on a model using time series data==

Executes a Kalman filter on a model, compliant with `KalmanFilterableProtocol`,
using time series observations from the input Databox. This method enables state
estimation and uncertainty quantification in line with the model's dynamics and
the time series data.

    kalman_output, output_info = self.kalman_filter(
        input_db,
        span,
        diffuse_scale=None,
        return_=("predict", "update", "smooth", "stds", "predict_err", "predict_mse_measurement", ),
        return_predict=True,
        return_update=True,
        return_smooth=True,
        return_stds=True,
        return_predict_err=True,
        return_predict_mse_measurement=True,
        rescale_variance=False,
        shocks_from_data=False,
        stds_from_data=False,
        prepend_initial=False,
        append_terminal=False,
        deviation=False,
        check_singularity=False,
        unpack_singleton=True
    )


### Input arguments ###


???+ input "self"
    The model, compliant with `KalmanFilterableProtocol`, performing the
    Kalman filtering.

???+ input "input_db"
    A Databox containing time series data to be used for filtering.

???+ input "span"
    A date span over which the filtering process is executed based on the
    measurement time series.

???+ input "diffuse_scale"
    A real number or `None`, specifying the scale factor for the diffuse
    initialization. If `None`, the default value is used.

???+ input "return_"
    An iterable of strings indicating which steps' results to return:
    "predict", "update", "smooth".

???+ input "return_predict"
    If `True`, return prediction step results.

???+ input "return_update"
    If `True`, return update step results.

???+ input "return_smooth"
    If `True`, return smoothing step results.

???+ input "rescale_variance"
    If `True`, rescale all variances by the optimal variance scale factor
    estimated using maximum likelihood after the filtering process.

???+ input "shocks_from_data"
    If `True`, use possibly time-varying shock values from the data; these
    values are interpreted as the medians (means) of the shocks. If `False`,
    zeros are used for all shocks.

???+ input "stds_from_data"
    If `True`, use possibly time-varying standard deviation values from the
    data. If `False`, currently assigned constant values are used for the
    standard deviations of all shocks.

???+ input "prepend_initial"
    If `True`, prepend observations to the resulting time series to cover
    initial conditions based on the model's maximum lag. No measurement
    observations are used in these initial time periods (even if some are
    available in the input data).

???+ input "append_terminal"
    If `True`, append observations to the resulting time series to cover
    terminal conditions based on the model's maximum lead. No measurement
    observations are used in these terminal time periods (even if some are
    available in the input data).

???+ input "deviation"
    If `True`, the constant vectors in transition and measurement equations are
    set to zeros, effectively running the Kalman filter as deviations from
    steady state (a balanced-growth path)

???+ input "check_singularity"
    If `True`, check the one-step ahead MSE matrix for the measurement variables
    for singularity, and throw a `SingularMatrixError` exception if the matrix
    is singular.

???+ input "unpack_singleton"
    If `True`, unpack `output_info` into a plain dictionary for models with a
    single variant.


### Returns ###


???+ returns "kalman_output"
    An object containing the following attributes, each being a Databox:

    | Attribute                  | Description
    |----------------------------|---------------------------------------------------
    | `predict_med`              | Medians from the prediction step.
    | `predict_std`              | Standard deviations from the prediction step.
    | `predict_mse_measurement`  | Mean squared error matrices from the prediction step.
    | `update_med`               | Medians from the update step.
    | `update_std`               | Standard deviations from the update step.
    | `predict_err`              | Prediction errors.
    | `smooth_med`               | Medians from the smoothing step.
    | `smooth_std`               | Standard deviations from the smoothing step.

    Some of these attributes may be `None` if the corresponding step was not
    requested in `return_`.

???+ returns "output_info"
    A dictionary containing additional information about the filtering process,
    such as log likelihood and variance scale. For models with multiple
    variants, `output_info` is a list of such dictionaries. If
    `unpack_singleton=False`, also `output_info` is a one-element list
    containing the dictionary for singleton models, too.

················································································
        """
        work_db = input_db.shallow()
        #if not shocks_from_data:
        #    shock_names = self.get_names(kind=_quantities.ANY_SHOCK, )
        #    work_db.remove(shock_names, strict_names=False, )
        #if not stds_from_data:
        #    std_names = self.get_names(kind=_quantities.ANY_STD, )
        #    work_db.remove(std_names, strict_names=False, )

        _LOGGER.set_level(logging_level, )

        num_variants \
            = self.resolve_num_variants_in_context(num_variants, )
        _LOGGER.debug(f"Running {num_variants} variants")

        slatable = self.get_slatable_for_kalman_filter(
            shocks_from_data=shocks_from_data,
            stds_from_data=stds_from_data,
        )

        input_ds = Dataslate.from_databox_for_slatable(
            slatable, work_db, span,
            num_variants=num_variants,
            prepend_initial=prepend_initial,
            append_terminal=append_terminal,
            clip_data_to_base_span=True,
        )

        frame = Frame(
            start=input_ds.periods[0],
            end=input_ds.periods[-1],
            simulation_end=input_ds.periods[-1],
        )

        needs = Needs(
            return_=return_,
            return_predict=return_predict,
            return_update=return_update,
            return_smooth=return_smooth,
            return_stds=return_stds,
            return_predict_err=return_predict_err,
            return_predict_mse_measurement=return_predict_mse_measurement,
            rescale_variance=rescale_variance,
        )

        squid = Squid(self, )
        qid_to_logly = self.create_qid_to_logly()

        logly_within_y = tuple(
            i for i, qid in enumerate(squid.y_qids, )
            if qid_to_logly.get(qid, False)
        )

        qid_to_name = self.create_qid_to_name()
        y_names = [ qid_to_name[qid] for qid in squid.y_qids ]
        name_to_log_name = {
            n: _quantities.wrap_logly(n, )
            for qid, n in enumerate(input_ds.names, )
            if qid_to_logly.get(qid, False, )
        } if qid_to_logly else {}

        output_store = _OutputStore(
            input_ds=input_ds,
            num_variants=0,
            measurement_names=y_names,
            needs=needs,
            name_to_log_name=name_to_log_name,
        )

        initialize = _ft.partial(
            _initializers.initialize,
            diffuse_method=diffuse_method,
            diffuse_scale=diffuse_scale,
        )

        zipped = zip(
            range(num_variants, ),
            self.iter_variants(),
            input_ds.iter_variants(),
        )

        output_info = []

        for vid, self_v, input_ds_v in zipped:

            solution_v = self_v.get_singleton_solution(deviation=deviation, )
            data_array = input_ds_v.get_data_variant()
            #
            # Initialize medians and stds
            #
            init_cov_u = self_v.get_cov_unanticipated_shocks()
            init_med, init_mse, unknown_init_impact \
                = initialize(solution_v, init_cov_u, )
            #
            # Presimulate the impact of anticipated shocks
            #
            all_v_impact \
                = _simulate_anticipated_shocks \
                (self_v, input_ds_v, frame, )
            #
            # Get values of observables
            #
            y1_array = data_array[squid.y_qids, :].copy()
            if logly_within_y:
                y1_array[logly_within_y, :] = _np.log(y1_array[logly_within_y, :])
            y1_incidence_array = ~_np.isnan(y1_array)
            #
            #
            output_store_v = _OutputStore(
                input_ds=input_ds_v,
                num_variants=1,
                measurement_names=y_names,
                needs=needs,
                name_to_log_name=name_to_log_name,
                transform=solution_v.Ua,
                squid=squid,
                measurement_incidence=y1_incidence_array,
            )

            generate_period_system = _ft.partial(
                _generate_period_system,
                solution_v=solution_v,
                y1_array=y1_array,
                std_u_array=data_array[squid.std_u_qids, :],
                std_w_array=data_array[squid.std_w_qids, :],
                all_v_impact=all_v_impact,
                initials=(init_med, init_mse, unknown_init_impact, ),
            )

            generate_period_data = _ft.partial(
                _generate_period_data,
                y_array=y1_array,
                u_array=data_array[squid.u_qids, :],
                v_array=data_array[squid.v_qids, :],
                w_array=data_array[squid.w_qids, :],
            )

            store_predict = _ft.partial(_OutputStore.store_predict, self=output_store_v, )
            store_update = _ft.partial(_OutputStore.store_update, self=output_store_v, )
            store_smooth = _ft.partial(_OutputStore.store_smooth, self=output_store_v, )

            cache = predict(
                num_periods=input_ds_v.num_periods,
                generate_period_system=generate_period_system,
                generate_period_data=generate_period_data,
                store_predict=store_predict,
                store_update=store_update,
                store_smooth=store_smooth,
                check_singularity=check_singularity,
            )

            if cache.needs_estimate_unknown_init:
                estimate_unknown_init(
                    cache=cache,
                )
                correct_for_unknown_init(
                    cache=cache,
                    store_predict=store_predict,
                )

            if needs.return_update:
                update(
                    cache=cache,
                    store_update=store_update,
                )

            if needs.return_smooth:
                smooth(
                    cache=cache,
                    store_smooth=store_smooth,
                )

            cache.calculate_likelihood(rescale_variance=needs.rescale_variance, )
            output_store_v.rescale_stds(cache.var_scale, )
            output_info_v = cache.create_output_info()

            output_store.extend(output_store_v, )
            output_info.append(output_info_v, )

        output_info = _has_variants.unpack_singleton(
            output_info, self.is_singleton,
            unpack_singleton=True,
        )

        return output_store.create_output_data(), output_info

    #]


class Cache:
    """
    """
    #[

    LOG_2_PI = _np.log(2 * _np.pi)

    _slots_to_input = (
        "num_periods",
    )

    _slots_to_preallocate = (
        "all_y",
        "all_num_obs",
        "all_Fi",
        "all_L",
        "all_a0",
        "all_y0",
        "all_u0",
        "all_v0",
        "all_w0",
        "all_Q0",
        "all_pe",
        "all_Zt_Fi",
        "all_T_G_prev",
        "all_H_cov_w",
        "all_P_cov_u",
        "all_Xi",
        "all_Z",
        "all_G",
        "all_M",
    )

    _other_slots = (
        "needs_estimate_unknown_init",
        "unknown_init_estimate",
        "sum_num_obs",
        "sum_log_det_F",
        "sum_pe_Fi_pe",
        "var_scale",
        "neg_log_likelihood",
        "neg_log_likelihood_contributions",
    )

    # __slots__ = _slots_to_input + _slots_to_preallocate + _other_slots

    def __init__(
        self: Self,
        **kwargs,
    ) -> None:
        """
        """
        for n in self._slots_to_input:
            setattr(self, n, kwargs[n], )
        for n in self._slots_to_preallocate:
            setattr(self, n, [None] * self.num_periods, )
        for n in self._other_slots:
            setattr(self, n, None, )

    def calculate_likelihood(self: Self, rescale_variance: bool, ) -> None:
        """
        """
        all_pe_Fi_pe = tuple(
            pe @ Fi @ pe if pe is not None and Fi is not None else None
            for pe, Fi in zip(self.all_pe, self.all_Fi, )
        )
        all_log_det_F = tuple(
            -_np.log(_np.linalg.det(Fi)) if Fi is not None else None
            for Fi in self.all_Fi
        )
        self.sum_num_obs = sum(i for i in self.all_num_obs if i is not None)
        self.sum_log_det_F = sum(i for i in all_log_det_F if i is not None)
        self.sum_pe_Fi_pe = sum(i for i in all_pe_Fi_pe if i is not None)
        self.var_scale = 1
        #
        if rescale_variance:
            self._calculate_variance_scale()
        #
        self.neg_log_likelihood_contributions = [
            (log_det_F + pe_Fi_pe + num_obs * self.LOG_2_PI)/2 if num_obs else 0
            for log_det_F, pe_Fi_pe, num_obs in zip(
                all_log_det_F, all_pe_Fi_pe, self.all_num_obs,
            )
            if num_obs is not None
        ]
        #
        self.neg_log_likelihood = (
            + self.sum_num_obs*self.LOG_2_PI
            + self.sum_log_det_F
            + self.sum_pe_Fi_pe
        ) / 2;

    def _calculate_variance_scale(self: Self, /, ) -> None:
        """
        """
        if self.sum_num_obs == 0:
            self.var_scale = 1
            self.sum_pe_Fi_pe = 0
            return
        self.var_scale = self.sum_pe_Fi_pe / self.sum_num_obs
        self.sum_log_det_F += self.sum_num_obs * _np.log(self.var_scale)
        self.sum_pe_Fi_pe = self.sum_pe_Fi_pe / self.var_scale

    def create_output_info(self: Self, /, ) -> dict[str, Any]:
        """
        """
        return {
            "neg_log_likelihood": self.neg_log_likelihood,
            "neg_log_likelihood_contributions": self.neg_log_likelihood_contributions,
            "var_scale": self.var_scale,
        }

    #]


def predict(
    num_periods: int,
    generate_period_system: Callable,
    generate_period_data: Callable,
    store_predict: Callable | None = None,
    store_update: Callable | None = None,
    store_smooth: Callable | None = None,
    check_singularity: bool = False,
) -> None:
    """
    """
    #[
    cache = Cache(num_periods=num_periods, )
    a1_prev, Q1_prev, Xi_prev = generate_period_system(-1, )
    needs_estimate_unknown_init = Xi_prev is not None
    cache.needs_estimate_unknown_init = needs_estimate_unknown_init
    cache.last_period_of_observations = -1
    #
    for t in range(num_periods, ):
        #
        T, P, K, Z, H, D, cov_u, cov_w, v_impact, *_ = generate_period_system(t, )
        y1, u0, v0, w0, *_ = generate_period_data(t, )
        #
        cache.all_y[t] = y1
        cache.all_num_obs[t] = y1.size
        any_y = cache.all_num_obs[t] > 0
        if any_y:
            cache.last_period_of_observations = t
        #
        if t > 0 and store_smooth:
            T_G_prev = T @ G_prev
            cache.all_T_G_prev[t-1] = T_G_prev
            cache.all_L[t-1] = T - T_G_prev @ Z_prev
        #
        # MSE prediction step
        #
        P_cov_u = P @ cov_u
        Q0 = T @ Q1_prev @ T.T + P_cov_u @ P.T
        Q0 = _make_symmetric(Q0)
        H_cov_w = H @ cov_w
        #
        F = Z @ Q0 @ Z.T + H_cov_w @ H.T \
            if any_y else _np.zeros((0, 0), dtype=_np.float64, )
        F = _make_symmetric(F)
        #
        # sing_values = _np.linalg.svd(F, compute_uv=False, )
        # relative_cond = sing_values[-1] / sing_values[0]
        # print(t, F.size, relative_cond, sing_values[0], )
        # TODO: Check if F singularity if requested by user
        if check_singularity and any_y:
            _check_singularity(F, t, )

        Fi = _np.linalg.inv(F) if any_y else _np.zeros((0, 0), )
        #
        # Median prediction step
        #
        a0 = T @ a1_prev + K + P @ u0 + (v_impact if v_impact is not None else 0)
        y0 = Z @ a0 + D + H @ w0
        #
        if store_predict:
            store_predict(t=t, a0=a0, y0=y0, u0=u0, v0=v0, w0=w0, Q0=Q0, F=F, cov_u0=cov_u, cov_w0=cov_w, )
        #
        # MSE updating step
        #
        Zt_Fi = Z.T @ Fi
        G = Q0 @ Zt_Fi
        Q1 = Q0 - G @ Z @ Q0
        Q1 = _make_symmetric(Q1)
        if store_update:
            store_update(t=t, Q=Q1, )
        #
        # Median updating step
        #
        pe = y1 - y0
        a1 = a0 + G @ pe
        #
        cache.all_a0[t] = a0
        cache.all_y0[t] = y0
        cache.all_pe[t] = pe
        cache.all_Fi[t] = Fi
        cache.all_Z[t] = Z
        #
        if store_smooth:
            cache.all_u0[t] = u0
            cache.all_v0[t] = v0
            cache.all_w0[t] = w0
            cache.all_Q0[t] = Q0
            cache.all_Zt_Fi[t] = Zt_Fi
            cache.all_P_cov_u[t] = P_cov_u
            cache.all_H_cov_w[t] = H_cov_w
        #
        if needs_estimate_unknown_init:
            Xi = ((T - T @ G_prev @ Z_prev) @ Xi_prev) if t > 0 else (T @ Xi_prev)
            cache.all_Xi[t] = Xi
            cache.all_G[t] = G
        #
        a1_prev = a1
        Q1_prev = Q1
        G_prev = G
        Z_prev = Z
        Xi_prev = Xi if needs_estimate_unknown_init else None
    #
    return cache
    #]


def estimate_unknown_init(
    cache: Cache,
    delta_tolerance: float = _DEFAULT_DELTA_TOLERANCE,
) -> None:
    """
    """
    #[
    all_M = tuple(
        (Z @ Xi) if Z is not None and Xi is not None else None
        for Z, Xi in zip(cache.all_Z, cache.all_Xi, )
    )
    all_Mt_Fi = tuple(
        (M.T @ Fi) if M is not None and Fi is not None else None
        for M, Fi in zip(all_M, cache.all_Fi, )
    )
    sum_Mt_Fi_M = sum(
        (Mt_Fi @ M) if Mt_Fi is not None and M is not None else 0
        for Mt_Fi, M in zip(all_Mt_Fi, all_M, )
    )
    sum_Mt_Fi_pe = sum(
        (Mt_Fi @ pe) if Mt_Fi is not None and pe is not None else 0
        for Mt_Fi, pe in zip(all_Mt_Fi, cache.all_pe, )
    )
    #
    sum_Mt_Fi_M = _make_symmetric(sum_Mt_Fi_M)
    sum_Mt_Fi_M[_np.abs(sum_Mt_Fi_M) < delta_tolerance] = 0
    #
    delta, *_ = _np.linalg.lstsq(sum_Mt_Fi_M, sum_Mt_Fi_pe, rcond=None, )
    #
    cache.unknown_init_estimate = delta
    cache.all_M = all_M
    #]


def correct_for_unknown_init(
    cache: Cache,
    store_predict: Callable | None,
) -> None:
    """
    """
    #[
    delta = cache.unknown_init_estimate
    for t in range(cache.num_periods, ):
        cache.all_a0[t] += cache.all_Xi[t] @ delta
        M_delta = cache.all_M[t] @ delta
        cache.all_y0[t] += M_delta
        cache.all_pe[t] -= M_delta
        if store_predict:
            store_predict(t=t, a0=cache.all_a0[t], y0=cache.all_y0[t], )
    #]


def update(
    cache: Cache,
    store_update: Callable | None,
) -> None:
    """
    """
    #[
    for t in range(cache.num_periods, ):
        a1, y1, u1, v1, w1, Q1, *_ = one_step_back(t, cache, )
        if store_update:
            pe = cache.all_pe[t]
            store_update(t=t, xi=a1, y=y1, u=u1, v=v1, w=w1, pe=pe, )
    #]


def smooth(
    cache: Cache,
    store_smooth: Callable | None,
) -> None:
    """
    """
    #[
    N = None
    r = None
    for t in reversed(range(cache.num_periods, )):
        a2, y2, u2, v2, w2, Q2, N, r = one_step_back(t, cache, N, r, )
        if store_smooth:
            store_smooth(t=t, xi=a2, y=y2, u=u2, v=v2, w=w2, Q=Q2, )
    #]


def one_step_back(
    t: int,
    cache: Cache,
    N: _np.ndarray | None = None,
    r: _np.ndarray | None = None,
    /,
) -> tuple:
    """
    """
    #[
    ak = cache.all_a0[t]
    y = cache.all_y[t]
    uk = cache.all_u0[t]
    vk = cache.all_v0[t]
    wk = cache.all_w0[t]
    Qk = cache.all_Q0[t]
    #
    if t <= cache.last_period_of_observations:
        T_G_prev = cache.all_T_G_prev[t]
        P_cov_u = cache.all_P_cov_u[t]
        H_cov_w = cache.all_H_cov_w[t]
        #
        a0 = cache.all_a0[t]
        Q0 = cache.all_Q0[t]
        L = cache.all_L[t]
        Zt_Fi = cache.all_Zt_Fi[t]
        Fi = cache.all_Fi[t]
        Z = cache.all_Z[t]
        pe = cache.all_pe[t]
        #
        Fi_pe = Fi @ pe
        Zt_Fi_pe = Zt_Fi @ pe
        Zt_Fi_Z = Zt_Fi @ Z
        #
        if N is None:
            N = Zt_Fi_Z
        else:
            N = Zt_Fi_Z + L.T @ N @ L
        Qk = Qk - Q0 @ N @ Q0
        Qk = _make_symmetric(Qk)
        #
        if r is None:
            wk = wk + H_cov_w.T @ Fi_pe
            r = Zt_Fi_pe
        else:
            wk = wk + H_cov_w.T @ (Fi_pe - (T_G_prev.T @ r))
            r = Zt_Fi_pe + L.T @ r
        ak = a0 + Q0 @ r
        uk = uk + P_cov_u.T @ r
        vk = vk
    #
    return ak, y, uk, vk, wk, Qk, N, r
    #]


def _generate_period_system(
    t: int,

    solution_v: Solution,
    y1_array: _np.ndarray,
    std_u_array: _np.ndarray,
    std_w_array: _np.ndarray,
    all_v_impact: Sequence[_np.ndarray | None],
    initials: tuple[_np.ndarray, _np.ndarray, int, ],
) -> tuple[_np.ndarray, ...]:
    """
    """
    #[
    if t == -1:
        return initials
    #
    T = solution_v.Ta
    P = solution_v.Pa
    K = solution_v.Ka
    inx_y = ~_np.isnan(y1_array[:, t], )
    Z = solution_v.Za[inx_y, :]
    H = solution_v.H[inx_y, :]
    D = solution_v.D[inx_y]
    cov_u = _np.diag(std_u_array[:, t]**2, )
    cov_w = _np.diag(std_w_array[:, t]**2, )
    v_impact = all_v_impact[t]
    return T, P, K, Z, H, D, cov_u, cov_w, v_impact,
    #]


def _generate_period_data(
    t,
    y_array: _np.ndarray,
    u_array: _np.ndarray,
    v_array: _np.ndarray,
    w_array: _np.ndarray,
) -> tuple[_np.ndarray, ...]:
    """
    """
    #[
    inx_y = ~_np.isnan(y_array[:, t], )
    y = y_array[inx_y, t]
    u = u_array[:, t]
    v = v_array[:, t]
    w = w_array[:, t]
    return y, u, v, w,
    #]


def _make_symmetric(matrix, ):
    return (matrix + matrix.T) / 2


def _check_singularity(F: _np.ndarray, t: int, ) -> None | NoReturn:
    rank_F = _np.linalg.matrix_rank(F)
    if rank_F < F.shape[0]:
        svd = _np.linalg.svd(F, )
        message = f"Singularity in prediction MSE matrix in period {t}, rank {rank_F} of {F.shape[0]}"
        raise SingularMatrixError(message, )




"""
Databox imports
"""


#[
from __future__ import annotations

from typing import (Self, )
from collections.abc import (Iterable, Callable, )
import csv as _cs
import numpy as _np
import dataclasses as _dc
import pickle as _pickle

from ..dates import (Period, Frequency, Span, )
from ..series.main import (Series, )
from .. import pages as _pages
#]


_DEFAULT_DATE_CREATOR = Period.from_sdmx_string


@_dc.dataclass
class _ImportBlock:
    """
    """
    #[
    row_index: Iterable[int] | None = None,
    periods: Iterable[Period] | None = None,
    column_start: int | None = None,
    num_columns: int | None = None,
    names: Iterable[str] | None = None,
    descriptions: Iterable[str] | None = None,

    def column_iterator(self, /, ):
        """
        """
        status = False
        names = self.names + [""]
        descriptions = self.descriptions + [""]
        current_columns = None
        current_name = None
        current_description = None
        for i, (n, d) in enumerate(zip(names, descriptions, )):
            if status and n!="*":
                status = False
                yield current_columns, current_name, current_description
            if status and n=="*":
                current_columns.append(i, )
            if not status and n and n!="*":
                status = True
                current_columns = [i]
                current_name = n
                current_description = d
    #]


class Inlay:
    """
    """
    #[
    @classmethod
    @_pages.reference(
        category="import_export",
        call_name="Databox.from_csv",
    )
    def from_csv(
        klass,
        file_name: str,
        *,
        date_creator: Callable | None = None,
        start_date_only: bool = False,
        description_row: bool = False,
        delimiter: str = ",",
        name_row_transform: Callable | None = None,
        csv_reader_settings: dict = {},
        numpy_reader_settings: dict = {},
        databox_settings: dict = {},
    ) -> Self:
        r"""
················································································


==Create a new Databox by reading time series from a CSV file==


    self = Databox.from_csv(
        file_name,
        *,
        date_creator=None,
        start_date_only=False,
        description_row=False,
        delimiter=",",
        csv_reader_settings={},
        numpy_reader_settings={},
        name_row_transform=None,
    )


### Input arguments ###


???+ input "file_name"
    Path to the CSV file to be read.

???+ input "date_creator"
    A callable for creating date objects from string representations. If `None`,
    a default method based on the SDMX string format is used.

???+ input "start_date_only"
    If `True`, only the start date of each time series is parsed from the CSV;
    subsequent periods are inferred based on frequency.

???+ input "description_row"
    Indicates if the CSV contains a row for descriptions of the time series.
    Defaults to `False`.

???+ input "delimiter"
    Character used to separate values in the CSV file.

???+ input "name_row_transform"
    A function to transform names in the name row of the CSV.

???+ input "csv_reader_settings"
    Additional settings for the CSV reader.

???+ input "numpy_reader_settings"
    Settings for reading data into numpy arrays.

???+ input "databox_settings"
    Settings for the Databox constructor.


### Returns ###


???+ returns "self"
    An `Databox` populated with time series from the CSV file.

················································································
        """
        self = klass(**databox_settings, )
        #
        num_header_rows = 1 + int(description_row)
        csv_rows = _read_csv(file_name, num_header_rows, **csv_reader_settings, )
        if not csv_rows:
            return self
        #
        header_rows = csv_rows[0:num_header_rows]
        data_rows = csv_rows[num_header_rows:]
        name_row = header_rows[0]
        #
        if name_row_transform:
            name_row = _apply_name_row_transform(name_row, name_row_transform, )
        #
        description_row = header_rows[1] if description_row else [""] * len(name_row)
        date_creator = date_creator or _DEFAULT_DATE_CREATOR
        #
        for b in _block_iterator(name_row, description_row, data_rows, date_creator, start_date_only, ):
            array = _read_array_for_block(file_name, b, num_header_rows, delimiter=delimiter, **numpy_reader_settings, )
            _add_series_for_block(self, b, array, )
        #
        return self

    @classmethod
    def from_sheet(klass, *args, **kwargs, ):
        """
        """
        return klass.from_csv(*args, **kwargs, )

    @classmethod
    def from_pickle(
        klass,
        file_name: str,
        /,
        **kwargs,
    ) -> Self:
        """
        """
        with open(file_name, "rb") as fid:
            return _pickle.load(fid, **kwargs, )

    #]


def _read_csv(file_name, num_header_rows, /, delimiter=",", **kwargs, ):
    """
    Read CSV cells into a list of lists
    """
    #[
    with open(file_name, "rt", encoding="utf-8-sig", ) as fid:
        all_rows = [ line for line in _cs.reader(fid, **kwargs, ) ]
    return all_rows
    #]


def _remove_nonascii_from_start(string, /, ):
    """
    Remove non-ascii characters from the start of a string
    """
    #[
    while string and not string[0].isascii():
        string = string[1:]
    return string
    #]


def _block_iterator(
    name_row,
    description_row,
    data_rows,
    date_creator,
    start_date_only,
    /,
):
    """
    """
    #[
    def _is_end(cell, /, ) -> bool:
        return cell.startswith("__")
    #
    def _is_start(cell, /, ) -> bool:
        if not cell.startswith("__"):
            return False
        try:
            letter = cell.removeprefix("__")[0]
            Frequency.from_letter(letter)
            return True
        except:
            return False
    #
    name_row += ["__"]
    status = False
    blocks = []
    current_date_column = None
    current_start = None
    current_frequency = None
    num_columns = len(name_row)
    for column, cell in enumerate(name_row):
        if status and _is_end(cell):
            status = False
            num_columns = column - current_start
            names = name_row[current_start:column]
            descriptions = description_row[current_start:column]
            row_index, periods = _extract_periods_from_data_rows(
                data_rows,
                current_frequency,
                current_date_column,
                date_creator,
                start_date_only,
            )
            yield _ImportBlock(row_index, periods, current_start, num_columns, names, descriptions)
        if not status and _is_start(cell):
            status = True
            current_date_column = column
            current_start = column + 1
            current_frequency = Frequency.from_letter(cell)
    #]


def _extract_periods_from_data_rows(
    data_rows,
    frequency: Frequency | None,
    column: int,
    date_creator: Callable,
    start_date_only: bool,
    /,
) -> tuple[tuple[int], tuple[Period]]:
    """
    """
    #[
    start_date = date_creator(data_rows[0][column], frequency=frequency, )
    date_extractor = {
        True: lambda i, line: start_date + i,
        False: lambda i, line: date_creator(line[column], frequency=frequency, ),
    }[start_date_only]
    row_indices_and_dates = ( 
        (i, date_extractor(i, line))
        for i, line in enumerate(data_rows)
        if start_date_only or line[column]
    )
    return tuple(zip(*row_indices_and_dates)) or ((), ())
    #]


def _read_array_for_block(file_name, block, num_header_rows, /, delimiter=",", **kwargs, ):
    #[
    skip_header = num_header_rows
    usecols = [ c for c in range(block.column_start, block.column_start+block.num_columns) ]
    return _np.genfromtxt(file_name, skip_header=skip_header, usecols=usecols, delimiter=delimiter, ndmin=2, **kwargs)
    #]


def _add_series_for_block(self, block, array, /, ):
    """
    """
    #[
    array = array[block.row_index, :]
    for columns, name, description in block.column_iterator():
        series = Series(num_variants=len(columns), description=description)
        series.set_data(block.periods, array[:, columns])
        self[name] = series
    #]

def _apply_name_row_transform(
    name_row: list[str],
    name_row_transform: Callable,
    /,
) -> list[str]:
    """
    """
    return [ name_row_transform(s) for s in name_row ]




# """
# Meta plans 
# """
# 
# 
# #[
# from __future__ import annotations
# 
# from collections.abc import (Iterable, )
# from typing import (Self, Any, Protocol, NoReturn, )
# from types import (EllipsisType, )
# import warnings as _wa
# import functools as _ft
# import numpy as _np
# import copy as _copy
# 
# from ..conveniences import copies as _copies
# from ..dates import (Period, )
# from .. import wrongdoings as _wrongdoings
# from .. import pages as _pages
# from . import _pretty as _pretty
# from . import _indexes as _indexes
# from . import transforms as _transforms
# #]
# 
# 
# __all__ = (
#     "SteadyPlan", "PlanSteady",
# )
# 
# 
# class SteadyPlannableProtocol(Protocol, ):
#     """
#     """
#     #[
# 
#     steady_can_be_exogenized: Iterable[str] | None
#     steady_can_be_endogenized: Iterable[str] | None
#     steady_can_be_fixed_level: Iterable[str] | None
#     steady_can_be_fixed_change: Iterable[str] | None
# 
#     #]
# 
# 
# @_pages.reference(
#     path=("structural_models", "simulation_plans.md", ),
#     categories={
#         "constructor": "Creating new simulation plans",
#         "definition_simultaneous": "Defining exogenized and endogenized data points in [`Simultaneous` simulations](simultaneous.md#simulate)",
#         "definition_sequential": "Defining exogenized and endogenized data points in [`Sequential` simulations](sequential.md#simulate)",
#         "information": "Getting information about simulation plans",
#         "information_simultaneous": "Getting information about simulation plans for [`Simultaneous` models](simultaneous.md)",
#         "information_sequential": "Getting information about simulation plans for [`Sequential` models](sequential.md)",
#     },
# )
# class SimulationPlan(
#     _pretty.PrettyMixin,
#     _indexes.ItemMixin,
#     _copies.CopyMixin,
# ):
#     """
# ················································································
# 
# Simulation meta plans
# ======================
# 
# `SimulationPlan` objects are used to set up meta information about
# conditioning assumptions for simulations of
# [`Simultaneous`](simultaneous_modelsd) or [`Sequential`](sequential_models)
# models. The simulation plans specify
# 
# * what variables to exogenize in what periods
# * what shocks to endogenized in what periods (`Simultaneous` models only)
# * what anticipation status to assign (`Simultaneous` models only)
# 
# The plans only contain meta information, not the actual data points for the
# exogenized variables. The actual data points are expected to be included in
# the input databox when the simulation is run.
# 
# ················································································
#     """
#     #[
# 
#     _registers = (
#         "exogenized",
#         "endogenized",
#         "exogenized_anticipated",
#         "exogenized_unanticipated",
#         "endogenized_unanticipated",
#         "endogenized_anticipated",
#     )
# 
#     __slots__ = (
#         ("base_span", )
#         + tuple(f"_can_be_{r}" for r in _registers)
#         + tuple(f"_{r}_register" for r in _registers)
#         + tuple(f"default_{r}" for r in _registers)
#     )
# 
#     @_pages.reference(
#         category="constructor",
#         call_name="SimulationPlan",
#     )
#     def __init__(
#         self,
#         model,
#         span: Iterable[Period] | None,
#     ) -> None:
#         """
# ················································································
# 
# ==Create new simulation plan object==
# 
# ```
# self = SimulationPlan(model, time_span, )
# ```
# 
# Create a new simulation plan object for a
# [`Simultaneous`](sequential_models) or
# [`Sequential`](sequential_models) model.
# 
# ### Input arguments ###
# 
# ???+ input "model"
# 
#     A [`Simultaneous`](sequential_models) or
#     [`Sequential`](sequential_models) model that will be simulated.
# 
# ???+ input "time_span"
# 
#     A date range on which the `model` will be simulated.
# 
# 
# ### Returns ###
# 
# ???+ returns "self"
# 
#     A new empty simulation plan object.
# 
# ················································································
#         """
#         self.base_span = tuple(span)
#         self._default_exogenized = None
#         self._default_endogenized = None
#         plannable = model.get_simulation_plannable()
#         for n in self._registers:
#             can_be_name = f"can_be_{n}"
#             register = {
#                 n: [None] * self.num_periods
#                 for n in getattr(plannable, can_be_name, )
#             } if hasattr(plannable, can_be_name, ) else {}
#             setattr(self, can_be_name, tuple(register.keys()))
#             setattr(self, f"_{n}_register", register)
# 
#     def check_consistency(
#         self,
#         plannable: SimulationPlannableProtocol,
#         span: Iterable[Period] | None,
#         /,
#     ) -> None:
#         """
#         """
#         benchmark = type(self)(plannable, span, )
#         if self.base_span != benchmark.base_span:
#             raise _wrongdoings.IrisPieError(f"Plan span must be the same as the simulation span")
#         for r in self._registers:
#             if getattr(self, f"can_be_{r}") != getattr(benchmark, f"can_be_{r}"):
#                 raise _wrongdoings.IrisPieError(f"Plan must be created using the simulated model")
# 
#     @property
#     @_pages.reference(category="property", )
#     def start(self, /, ) -> Period:
#         """==Start date of the simulation span=="""
#         return self.base_span[0]
# 
#     @property
#     @_pages.reference(category="property", )
#     def end(self, /, ) -> Period:
#         """==End date of the simulation span=="""
#         return self.base_span[-1]
# 
#     @property
#     @_pages.reference(category="property", )
#     def num_periods(self, /, ) -> int:
#         """==Number of periods in the simulation span=="""
#         return len(self.base_span) if self.base_span is not None else 1
# 
#     @property
#     @_pages.reference(category="property", )
#     def frequency(self, /, ) -> str:
#         """==Date frequency of the simulation span=="""
#         return self.start.frequency
# 
#     @_pages.reference(category="definition_sequential", )
#     def exogenize(
#         self,
#         dates: Iterable[Period] | EllipsisType,
#         names: Iterable[str] | str | EllipsisType,
#         /,
#         *,
#         transform: str | None = None,
#         # when_data: bool | None = None,
#         **kwargs,
#     ) -> None:
#         r"""
# ................................................................................
# 
# ==Exogenize certain LHS quantities at certain dates==
# 
# Exogenize certain LHS quantities at specified dates, setting them as
# predetermined values within the simulation of
# a [`Sequential` model](sequential.md). This method is used to control how
# the model behaves during simulations by fixing certain variables to known
# values.
# 
#     self.exogenize(
#         dates,
#         names,
#         *,
#         transform=None,
#         when_data=False,
#     )
# 
# ### Input arguments ###
# 
# ???+ input "self"
#     The simulation plan in which data points will be exogenized.
# 
# ???+ input "dates"
#     A list of dates or `...` to apply to all dates at which the quantities 
#     will be exogenized.
# 
# ???+ input "names"
#     A list of names or a single name, or `...` to apply to all names that 
#     specifies which quantities to set as predetermined at the specified dates.
# 
# ???+ input "transform"
#     Specifies the transformation to apply to the exogenized quantities. If not
#     specified, no transformation is applied. Available transformations include:
# 
#     * `None`: Exogenize the LHS variables as they are with no
#     transformation.
# 
#     * `"log"`: Exogenize the natural logarithm of the LHS variables. Input
#     time series needs to be prefixed with `log_`.
# 
#     * `"diff"`: Exogenize the first difference of the LHS variables. Input
#     time series needs to be prefixed with `diff_`.
# 
#     * `"diff_log"`: Exogenize the first difference of the natural logarithm
#     of the LHS variables. Input time series needs to be prefixed with
#     `diff_log_`.
# 
#     * `"roc"`: The gross rate of change of the LHS variables from one
#     period to the next. Input time series needs to be prefixed with `roc_`.
# 
#     * `"pct"`: The percentage change of the LHS variables from one period
#     to the next. Input time series needs to be prefixed with `pct_`.
# 
# ???+ input "when_data"
#     Specifies whether the exogenization should only occur if a valid 
#     value exists in the input data.
# 
# ### Returns ###
# 
# This method modifies the object in place and does not return a value.
# 
# ................................................................................
#         """
#         transform = _transforms.resolve_transform(transform, **kwargs, )
#         self._write_to_register("exogenized", dates, names, transform, )
# 
#     @_pages.reference(category="definition_simultaneous", )
#     def exogenize_anticipated(
#         self,
#         dates: Iterable[Period] | EllipsisType,
#         names: Iterable[str] | str | EllipsisType,
#         *,
#         status: bool | int = True,
#     ) -> None:
#         """
# ················································································
# 
# ==Exogenize certain quantities at certain dates==
# 
# ```
# self.exogenize_anticipated(
#     dates,
#     names,
# )
# ```
# 
# ### Input arguments ###
# 
# 
# ???+ input "dates"
# 
#     Dates at which the `names` will be exogenized; use `...` for all simulation dates.
# 
# ???+ input "names"
# 
#     Names of quantities to exogenize at the `dates`; use `...` for all exogenizable quantities.
# 
# ················································································
#         """
#         self._write_to_register(
#             "exogenized_anticipated",
#             dates,
#             names,
#             status,
#         )
# 
#     @_pages.reference(category="definition_simultaneous", )
#     def exogenize_unanticipated(
#         self,
#         dates: Iterable[Period] | EllipsisType,
#         names: Iterable[str] | str | EllipsisType,
#         *,
#         status: bool | int = True,
#     ) -> None:
#         r"""
# ················································································
# 
# ==Exogenize certain quantities at certain dates as unanticipated==
# 
# ```
# self.exogenize_unanticipated(
#     dates, names,
#     /,
#     transform=None,
#     when_data=False,
# )
# ```
# 
# ### Input arguments ###
# 
# 
# ???+ input "dates"
# 
#     Dates at which the `names` will be exogenized; use `...` for all simulation dates.
# 
# ???+ input "names"
# 
#     Names of quantities to exogenize at the `dates`; use `...` for all exogenizable quantities.
# 
# 
# ### Input arguments available only for `Sequential` models ###
# 
# ???+ input "transform"
# 
#     Transformation (specified as a string) to be applied to the exogenized
#     quantities; if `None`, no tranformation is applied.
# 
# ???+ input "when_data"
# 
#     If `True`, the data point will be exogenized only if a proper value
#     exists in the input data.
# 
# ················································································
#         """
#         self._write_to_register(
#             "exogenized_unanticipated",
#             dates,
#             names,
#             status,
#         )
# 
#     @_pages.reference(category="information_sequential", )
#     def get_exogenized_in_period(self, *args, **kwargs, ) -> tuple[str, ...]:
#         """
# ················································································
# 
# ==Get names exogenized in a certain period==
# 
# ················································································
#         """
#         return self._get_names_registered_in_period(
#             self._exogenized_register,
#             *args, **kwargs,
#         )
# 
#     @_pages.reference(category="information_simultaneous", )
#     def get_exogenized_unanticipated_in_period(self, *args, **kwargs, ) -> tuple[str, ...]:
#         """
# ················································································
# 
# ==Get names exogenized as unanticipated in a certain period==
# 
# ················································································
#         """
#         return self._get_names_registered_in_period(
#             self._exogenized_unanticipated_register,
#             *args, **kwargs,
#         )
# 
#     @_pages.reference(category="information_simultaneous", )
#     def get_exogenized_anticipated_in_period(self, *args, **kwargs, ) -> tuple[str, ...]:
#         """
# ················································································
# 
# ==Get names exogenized as anticipated in a certain period==
# 
# ················································································
#         """
#         return self._get_names_registered_in_period(
#             self._exogenized_anticipated_register,
#             *args, **kwargs,
#         )
# 
#     # @_pages.reference(category="definition_sequential", )
#     def endogenize(
#         self,
#         dates: Iterable[Period] | EllipsisType,
#         names: Iterable[str] | str | EllipsisType,
#         /,
#     ) -> None:
#         r"""
#         """
#         self._write_to_register("endogenized", dates, names, True, )
# 
#     @_pages.reference(category="definition_simultaneous", )
#     def endogenize_anticipated(
#         self,
#         dates: Iterable[Period] | EllipsisType,
#         names: Iterable[str] | str | EllipsisType,
#         *,
#         status: bool | int = True,
#     ) -> None:
#         """
# ················································································
# 
# ==Endogenize certain quantities at certain dates==
# 
# ················································································
#         """
#         self._write_to_register(
#             "endogenized_anticipated",
#             dates,
#             names,
#             status,
#         )
# 
#     @_pages.reference(category="definition_simultaneous", )
#     def endogenize_unanticipated(
#         self,
#         dates: Iterable[Period] | EllipsisType,
#         names: Iterable[str] | str | EllipsisType,
#         *,
#         status: bool | int = True,
#     ) -> None:
#         """
# ················································································
# 
# ==Endogenize certain quantities at certain dates==
# 
# ················································································
#         """
#         self._write_to_register(
#             "endogenized_unanticipated",
#             dates,
#             names,
#             status,
#         )
# 
#     @_pages.reference(category="information_simultaneous", )
#     def get_endogenized_unanticipated_in_period(self, *args, **kwargs, ) -> tuple[str, ...]:
#         """
# ················································································
# 
# ==Get names endogenized as unanticipated in a certain period==
# 
# ················································································
#         """
#         return self._get_names_registered_in_period(
#             self._endogenized_unanticipated_register,
#             *args,
#             **kwargs,
#         )
# 
#     @_pages.reference(category="information_simultaneous", )
#     def get_endogenized_anticipated_in_period(self, *args, **kwargs, ) -> tuple[str, ...]:
#         """
# ················································································
# 
# ==Get names endogenized as anticipated in a certain period==
# 
# ················································································
#         """
#         return self._get_names_registered_in_period(
#             self._endogenized_unanticipated_register,
#             *args,
#             **kwargs,
#         )
# 
#     @property
#     def needs_split(self, /, ) -> bool:
#         r"""
#         True if there is any unanticipated point in the plan in the second
#         or later simulation period
#         """
#         return any(
#             any(
#                 any(_is_active_statue(i) for i in v[1:])
#                 for v in getattr(self, "_" + a + "_register").values()
#             )
#             for a in ("exogenized_unanticipated", "endogenized_unanticipated", )
#         )
# 
#     @property
#     def is_empty(self, /, ) -> bool:
#         """
#         True if there are no exogenized or endogenized points in the plan
#         """
#         has_any_points = any(
#             _has_points_in_register(self._get_register_by_name(n, ), )
#             for n in self._registers
#         )
#         return not has_any_points
# 
#     # def swap(
#     #     self,
#     #     dates: Iterable[Period] | EllipsisType,
#     #     pairs: Iterable[tuple[str, str]] | tuple[str, str],
#     #     *args, **kwargs,
#     # ) -> None:
#     #     """
#     #     """
#     #     pairs = tuple(pairs)
#     #     if not pairs:
#     #         return
#     #     if len(pairs) == 2 and isinstance(pairs[0], str) and isinstance(pairs[1], str):
#     #         pairs = (pairs, )
#     #     for pair in pairs:
#     #         self.exogenize(dates, pair[0], *args, **kwargs, )
#     #         self.endogenize(dates, pair[1], *args, **kwargs, )
# 
#     def swap_anticipated(
#         self,
#         dates: Iterable[Period] | EllipsisType,
#         pairs: Iterable[tuple[str, str]] | tuple[str, str],
#         *args, **kwargs,
#     ) -> None:
#         r"""
# ................................................................................
# 
# ==Swap quantities as anticipated at certain dates==
# 
# Swap (exogenize and endogenize) quantities at certain dates. This method
# exogenizes the first quantity in the pair and endogenizes the second
# quantity in the pair at the specified dates. It is equivalent to calling
# `exogenize_anticipated` and `endogenize_anticipated` separately.
# 
#     self.swap_anticipated(
#         dates,
#         pairs,
#     )
# 
# 
# ### Input arguments ###
# 
# ???+ input "self"
#     The simulation plan in which data points will be exogenized and
#     endogenized.
# 
# ???+ input "dates"
#     Dates at which the quantities will be exogenized and endogenized.
# 
# ???+ input "pairs"
#     A list of pairs of names to exogenize and endogenize at the specified
#     dates.
# 
# 
# ### Returns ###
# 
# Returns no value; the method modifies the `SimulationPlan` object in place.
# 
# ................................................................................
#         """
#         pairs = tuple(pairs)
#         if not pairs:
#             return
#         if len(pairs) == 2 and isinstance(pairs[0], str) and isinstance(pairs[1], str):
#             pairs = (pairs, )
#         for pair in pairs:
#             self.exogenize_anticipated(dates, pair[0], *args, **kwargs, )
#             self.endogenize_anticipated(dates, pair[1], *args, **kwargs, )
# 
#     def swap_unanticipated(
#         self,
#         dates: Iterable[Period] | EllipsisType,
#         pairs: Iterable[tuple[str, str]] | tuple[str, str],
#         *args, **kwargs,
#     ) -> None:
#         """
# ................................................................................
# 
# ==Swap quantities as unanticipated at certain dates==
# 
# Swap (exogenize and endogenize) quantities at certain dates. This method
# exogenizes the first quantity in the pair and endogenizes the second
# quantity in the pair at the specified dates. It is equivalent to calling
# `exogenize_unanticipated` and `endogenize_unanticipated` separately.
# 
#     self.swap_unanticipated(
#         dates,
#         pairs,
#     )
# 
# 
# ### Input arguments ###
# 
# ???+ input "self"
#     The simulation plan in which data points will be exogenized and
#     endogenized.
# 
# ???+ input "dates"
#     Dates at which the quantities will be exogenized and endogenized.
# 
# ???+ input "pairs"
#     A list of pairs of names to exogenize and endogenize at the specified
#     dates.
# 
# 
# ### Returns ###
# 
# Returns no value; the method modifies the `SimulationPlan` object in place.
# 
# ................................................................................
#         """
#         pairs = tuple(pairs)
#         if not pairs:
#             return
#         if len(pairs) == 2 and isinstance(pairs[0], str) and isinstance(pairs[1], str):
#             pairs = (pairs, )
#         for pair in pairs:
#             self.exogenize_unanticipated(dates, pair[0], *args, **kwargs, )
#             self.endogenize_unanticipated(dates, pair[1], *args, **kwargs, )
# 
#     def get_register_as_bool_array(
#         self: Self,
#         register_name: str,
#         names: str | Iterable[str] | EllipsisType = ...,
#         periods: Iterable[Period] | EllipsisType = ...,
#     ) -> _np.ndarray:
#         """
#         """
#         register = self._get_register_by_name(register_name, )
#         per_indexes = self._get_per_indexes(periods, )
#         names = _resolve_validate_register_names(register, names, register_name, )
#         num_names, num_pers = len(names), len(per_indexes)
#         #
#         if names and per_indexes:
#             def get_points_for_name(name: str, ) -> tuple[bool, ...]:
#                 return tuple(
#                     register[name][t] if t is not None else False
#                     for t in per_indexes
#                 )
#             array = _np.array(tuple(
#                 get_points_for_name(n, ) for n in names
#             ), dtype=bool, )
#         else:
#             array = _np.zeros((num_names, num_pers, ), dtype=bool, )
#         #
#         return array
# 
#     def _get_per_indexes(
#         self,
#         periods: Iterable[Period] | EllipsisType,
#         /,
#     ) -> tuple[int | None, ...]:
#         """
#         """
#         if periods is ...:
#             return tuple(range(self.num_periods, ), )
#         else:
#             return tuple(
#                 t - self.start if self._is_per_in_span(t, ) else None
#                 for t in periods
#             )
# 
#     def _is_per_in_span(self, per: Period, ) -> bool:
#         return per >= self.start and per <= self.end
# 
#     def _get_register_by_name(self, name: str, /, ) -> dict[str, Any]:
#         """
#         """
#         return getattr(self, f"_{name}_register", )
# 
#     def get_exogenized_point(
#         self,
#         name: str,
#         date: Period,
#         /,
#     ) -> _transforms.PlanTransformProtocol | None:
#         """
#         """
#         column = next((
#             column
#             for column, t in enumerate(self.base_span)
#             if date == t
#         ))
#         point = self._exogenized_register[name][column]
#         return point if point is not None else self._default_exogenized
# 
#     def get_endogenized_point(
#         self,
#         name: str,
#         column: int,
#         /,
#     ) -> _transforms.PlanTransformProtocol | None:
#         """
#         """
#         point = self._endogenized_register[name][column]
#         return point if point is not None else self._default_endogenized
# 
#     def get_databox_names(self, /, ) -> tuple[str]:
#         """
#         """
#         databox_names = set()
#         for k, v in self._exogenized_register.items():
#             databox_names.update(
#                 t.resolve_databox_name(k, )
#                 for t in v if t is not None
#             )
#         return tuple(n for n in databox_names if n is not None)
# 
#     def _get_names_registered_in_period(
#         self,
#         register: dict[str, Any],
#         date: Period,
#     ) -> tuple[str, ...]:
#         """
#         """
#         column_index = self.base_span.index(date)
#         return tuple(
#             name
#             for name, status in register.items()
#             if bool(status[column_index])
#         )
# 
#     def __str__(self, /, ) -> str:
#         """
#         """
#         return self.get_pretty_string()
# 
#     def _write_to_register(
#         self,
#         register_name: str,
#         periods: Iterable[Period] | EllipsisType,
#         names: Iterable[str] | str | EllipsisType,
#         new_status: Any,
#     ) -> None:
#         """
#         """
#         register = self._get_register_by_name(register_name, )
#         names = _resolve_validate_register_names(register, names, register_name, )
#         per_indexes, *_ = self._get_period_indexes(periods, )
#         for n in names:
#             for t in per_indexes:
#                 register[n][t] = new_status
# 
#     def _get_period_indexes(
#         self,
#         periods: Iterable[Period] | EllipsisType,
#         /,
#     ) -> tuple[tuple[int, ...], tuple[Period, ...]]:
#         """
#         """
#         if periods is Ellipsis:
#             periods = tuple(self.base_span)
#         if hasattr(periods, "resolve"):
#             periods = periods.resolve(self, )
#         catch_invalid_periods(periods, self.base_span, )
#         period_indexes = tuple(d - self.start for d in periods)
#         return period_indexes, periods
# 
#     #]
# 
# 
# @_pages.reference(
#     path=("structural_models", "steady_plans.md", ),
#     categories={
#         "constructor": "Creating new steady plans",
#         "property": None,
#         "definition": "Defining Exogenized, endogenized and fixed quantities",
#     },
# )
# class SteadyPlan:
#     """
# ················································································
# 
# Steady-state meta plans
# ========================
# 
# `SteadyPlan` objects set up meta assumptions about the steady state values of
# certain model quantities (variables and parameters), either for reverse
# engineered calculations or for narrowing down the underdetermined
# steady-state systems.
# 
# ················································································
#     """
#     #[
# 
#     __slots__ = (
#         "can_be_exogenized",
#         "can_be_endogenized",
#         "can_be_fixed_level",
#         "can_be_fixed_change",
#         "_exogenized_register",
#         "_endogenized_register",
#         "_fixed_level_register",
#         "_fixed_change_register",
#     )
# 
#     @_pages.reference(category="constructor", call_name="SteadyPlan", )
#     def __init__(
#         self,
#         model,
#         /,
#     ) -> None:
#         """
# ················································································
# 
# ==Create new steady plan object==
# 
# ················································································
#         """
#         plannable = model.get_steady_plannable()
#         for r in ("exogenized", "endogenized", "fixed_level", "fixed_change", ):
#             can_be_name = f"can_be_{r}"
#             register = {
#                 n: False
#                 for n in getattr(plannable, can_be_name, )
#             } if hasattr(plannable, can_be_name, ) else {}
#             setattr(self, can_be_name, tuple(register.keys()))
#             setattr(self, f"_{r}_register", register)
# 
#     #]
# 
# #         self.exogenized = {
# #             n: [None] * self.num_periods
# #             for n in plannable.can_be_exogenized
# #         } if plannable.can_be_exogenized else {}
# #         #
# #         self.endogenized = {
# #             n: [None] * self.num_periods
# #             for n in plannable.can_be_endogenized
# #         } if plannable.can_be_endogenized else {}
# #         #
# #         self.fixed_level = {
# #             n: [None] * self.num_periods
# #             for n in plannable.can_be_fixed_level
# #         } if plannable.can_be_fixed_level else {}
# #         #
# #         self.fixed_change = {
# #             n: [None] * self.num_periods
# #             for n in plannable.can_be_fixed_change
# #         } if plannable.can_be_fixed_change else {}
# 
#     @_pages.reference(category="definition", )
#     def exogenize(
#         self,
#         names: Iterable[str] | str,
#         /,
#     ) -> None:
#         """
# ················································································
# 
# ==Exogenize steady levels of variables==
# 
# ················································································
#         """
#         pass
# 
#     @_pages.reference(category="definition", )
#     def endogenize(
#         self,
#         names: Iterable[str] | str,
#         /,
#     ) -> None:
#         """
# ················································································
# 
# ==Endogenize parameters==
# 
# ················································································
#         """
#         pass
# 
#     @_pages.reference(category="definition", )
#     def fix_level(
#         self,
#         names: Iterable[str] | str,
#         /,
#     ) -> None:
#         """
# ················································································
# 
# ==Fix steady levels of variables==
# 
# ················································································
#         """
#         pass
# 
#     @_pages.reference(category="definition", )
#     def fix_change(
#         self,
#         names: Iterable[str] | str,
#         /,
#     ) -> None:
#         """
# ················································································
# 
# ==Fix steady changes of variables==
# 
# ················································································
#         """
#         pass
# 
# 
# def _resolve_validate_register_names(
#     register: dict | None,
#     names: Iterable[str] | str | EllipsisType,
#     register_name: str,
#     /,
# ) -> tuple[str]:
#     """
#     """
#     keys = tuple(register.keys()) if register else ()
#     if names is Ellipsis:
#         return keys
#     names = tuple(names) if not isinstance(names, str) else (names, )
#     invalid = tuple(n for n in names if n not in keys)
#     if invalid:
#         message = (f"These names cannot be {register_name}:", ) + invalid
#         raise _wrongdoings.IrisPieCritical(message, )
#     return names
# 
# 
# def catch_invalid_periods(
#     dates: Iterable[Period],
#     base_span: tuple[Period],
#     /,
# ) -> NoReturn | None:
#     """
#     """
#     invalid = tuple(repr(d) for d in dates if d not in base_span)
#     if invalid:
#         raise _wrongdoings.IrisPieError(
#             ("These date(s) are out of simulation span:", ) + invalid
#         )
# 
# 
# def _is_active_statue(value: Any, /, ) -> bool:
#     return (value is not None) and (value is not False)
# 
# 
# def _has_points_in_register(register: dict, /, ) -> bool:
#     return any(
#         any(_is_active_statue(i) for i in v)
#         for v in register.values()
#     )
# 
# 
# PlanSteady = SteadyPlan
# 



"""
Implement SimulationPlannableProtocol and SteadyPlannableProtocol
"""


#[
from __future__ import annotations

from .. import quantities as _quantities
#]


class _SimulationPlannable:
    """
    """
    #[

    def __init__(
        self,
        simultaneous,
        **kwargs,
    ) -> None:
        """
        """
        quantities = simultaneous.quantities
        solution_vectors = simultaneous.solution_vectors
        #
        curr_xi_qids, *_ = solution_vectors.get_curr_transition_indexes()
        qid_to_name = simultaneous.create_qid_to_name()
        can_be_exogenized = tuple( qid_to_name[qid] for qid in curr_xi_qids )
        #
        self.can_be_exogenized_anticipated = can_be_exogenized
        self.can_be_exogenized_unanticipated = can_be_exogenized
        #
        self.can_be_endogenized_anticipated \
            = tuple(_quantities.generate_quantity_names_by_kind(
                quantities, _quantities.ANTICIPATED_SHOCK,
            ))
        #
        self.can_be_endogenized_unanticipated \
            = tuple(_quantities.generate_quantity_names_by_kind(
                quantities, _quantities.UNANTICIPATED_SHOCK,
            ))


class Inlay:
    """
    """
    #[

    def get_simulation_plannable(self, **kwargs, ) -> _SimulationPlannable:
        """
        """
        return _SimulationPlannable(self, **kwargs, )

    #]




"""
"""


#[
from __future__ import annotations

from typing import (Any, Iterable, )
import wlogging as _wl

from ..dates import (Period, )
from ..dataslates.main import (Dataslate, )
#]


_LOGGER = _wl.get_colored_two_liner(__name__, level=_wl.INFO, )


_Info = dict[str, Any] | list[dict[str, Any]]


class Mixin:
    """
    """
    #[

    def multiply_stds(
        self,
        std_db: Dataslate,
        multiplier_db: Dataslate,
        span: Iterable[Period],
        /,
        *,
        num_variants: int | None = None,
        logging_level: int = _wl.INFO,
        unpack_singleton: bool = True,
        **kwargs,
    ) -> tuple[Dataslate, _Info]:
        """
        """
        _LOGGER.set_level(logging_level, )

        num_variants \
            = self.resolve_num_variants_in_context(num_variants, )

        base_dates = tuple(span, )

        std_slatable, multiplier_slatable = self.get_slatables_for_multiply_stds()

        std_dataslate = Dataslate.from_databox_for_slatable(
            std_slatable, std_db, base_dates,
            num_variants=num_variants,
        )

        multiplier_dataslate = Dataslate.from_databox_for_slatable(
            multiplier_slatable, multiplier_db, base_dates,
            num_variants=num_variants,
        )

        zipped = zip(
            range(num_variants),
            std_dataslate.iter_variants(),
            multiplier_dataslate.iter_variants(),
        )

        #=======================================================================
        # Main loop over variants
        #
        output_info = []
        for vid, std_ds_v, multiplier_ds_v in zipped:
            output_info_v = {}
            #
            std_array = std_ds_v.get_data_variant()
            multiplier_array = multiplier_ds_v.get_data_variant()
            names = std_ds_v._invariant.names
            #
            zipped = zip(names, std_array, multiplier_array)
            #
            for names, std_row, multiplier_row in zipped:
                std_row *= multiplier_row
            #
            output_info.append(info_v, )
            #
        #=======================================================================

        output_db = std_dataslate.to_databox()

        if target_databox is not None:
            output_db = target_databox | output_db

        is_singleton = num_variants == 1
        output_info = _has_variants.unpack_singleton(
            output_info, is_singleton,
            unpack_singleton=unpack_singleton,
        )
        return output_db, output_info

    #]

